from __future__ import annotations
from typing import List
import sys, time,pydevd
from PySide2.QtCore import Qt, QObject, QThread, Signal, Slot
from Project.ProjectSettings import ProjectSettings
from Frames.FrameParser import FrameParser
from Frames.Frame import Frame
from Frames.EbdFrame import EbdFrame
from Frames.FrameAddress import FrameAddress
from Utilities.Constants import Constants
from Utilities.Log import Log
from UI.UISignals import UISignals
from Communication.CommandManager import CommandManager
from Communication.SerialPort import SerialPort
import random

class UserApplicationSignals(QObject):
    """This class holds the signals user application
    """
    #: Signal to be called to stop the user application      
    ApplicationStopped = Signal(bool)

class UserApplication(QThread):
    """This class can be used as user application placeholder which could be able to be loaded and executed dynamically
    """

    # #: The serial port to be used by :class:SerialPort
    # __SERIAL_PORT = 'COM9'
    # #: The baud rate for :class:SerialPort
    # __SERIAL_PORT_BAUDRATE = 115200

    def __init__(self, project : ProjectSettings, frameParser : FrameParser, remoteDeviceSwIdReceived : QObject, parent=None):  
        """Class constructor

        :param project: The project settings
        :type project: ProjectSettings
        :param frameParser: The :class:FrameParser as created by the caller
        :type frameParser: FrameParser
        :param remoteDeviceSwIdReceived: Callbacl function to be called when a SW version and ID are received
        :type remoteDeviceSwIdReceived: QObject
        :param parent: The parent object, defaults to None
        :type parent: [type], optional
        """                  
        super().__init__(parent)
        self._execute = False
        self._project = project
        self._frameParser = frameParser
        self._commandManager = None
        self._uiSignals = UISignals()
        self._uiSignals.RemoteDeviceSwIdReceived.connect(remoteDeviceSwIdReceived)

    def Exit(self):
        """Exits the user application
        This method terminates the thread and the closes the :class:CommandManager
        """        
        self.StopClicked()
        try:
            if self._commandManager:
                self._commandManager.Close()              
        except Exception as e:
            Log.PrintException(f'UserApplication.Exit: {str(e)}') 

    @Slot(bool)
    def StopClicked(self, value : bool = True):
        """Stops the thread

        :param value: User provided value (from the parent object), defaults to True
        :type value: bool, optional
        """        
        self._execute = False
        if self.isRunning():
            self.terminate()

    def run(self):
        """This method holds the user implementation executed by this thread
        """
        # pydevd is required for debugging which can be installed with .. code-block:: python pip install pydevd        
        #pydevd.connected = True
        #pydevd.settrace(suspend=False)

        try:
            self._commandManager = CommandManager(self._project.FpgaDevice, self._project.IpAddress, self._project.TcpPort)
            self._execute = True

            # Read S/N ans SW version - use allways this mwthod to inform the main application about SW version and remote device S/N
            boardInfo = self._commandManager.ReadId()
            serialNumberHex = '{0:0{1}X}'.format(boardInfo.SerialNumber, 8)
            self._uiSignals.RemoteDeviceSwIdReceived.emit(f'{serialNumberHex} : {boardInfo.Version}')


            # Configure the FPGA - Configure the FPGA just once 
            status = self._commandManager.ConfigureDevice(self._project.BitFilePath)
            Log.Print(str(status))

            
            
            #Extract fault list
            #Finds all the sensitive bits which are non-masked
            # :return: A list of tuples where each tuple consists of:
            # +-------+------------------------+-----------------------------------------------------------------+
            # | Index | Name                   | Description                                                     |
            # +=======+========================+=================================================================+
            # | 0     | frameIndex             | The index in the provided list where the item resides           |
            # +-------+------------------------+-----------------------------------------------------------------+
            # | 1     | bitIndex               | The bit index (which is sensitive and non-masked) in the frame  |
            # +-------+------------------------+-----------------------------------------------------------------+
            # | 2     | frame address value    | The frame address                                               |
            # +-------+------------------------+-----------------------------------------------------------------+            
            nonMaskedSensitiveBits = EbdFrame.FindNonMaskedSensitiveBits(self._frameParser.EbdFrames, self._frameParser.MskFrames)
        
            
            #Location (frames) of the PBLOCK/DUT we will inject faults
            frames=[]
            

            f=open(self._project.Path+"/frames.txt","r")
            for line in f:
                tmp=line.replace("\n","")
                frm=int(tmp, 16)
                frames.append(frm)
            f.close()


            #Bit locations of non masked bits
            dut_non_masked_bits=[]
            for item in nonMaskedSensitiveBits:
               if item[2] in frames:
                   dut_non_masked_bits.append(item)
            

            ############################# Fault injection procedure ################################

            # Preconditions: Set UPSETS_PER_INJECTION, DIFF_INPUTS_TO_TEST and INJECTIONS_TO_PERFROM
            # Repeat the following procedure for #INJECTIONS_TO_PERFROM times
            #   - Inject a fault (SBU or MBU --> see #UPSETS_PER_INJECTION) into a non masked bit of PBLOCK/DUT
            #   - Check if the fault resulted in an error by providing #DIFF_INPUTS_TO_TEST to the DUT and comparing its output result with a golden reference 
            # Post-processing: AVF is the total_errors / INJECTIONS_TO_PERFROM

            #Number of bit flips per test
            UPSETS_PER_INJECTION=1

            #Define the number of different inputs to check for every fault
            DIFF_INPUTS_TO_TEST=100

            #Number of errors
            total_errors=0

            INJECTIONS_TO_PERFROM=100

            
            for i in range(INJECTIONS_TO_PERFROM):
                
                rand_items=random.choices(dut_non_masked_bits,k=UPSETS_PER_INJECTION)

                #Get the frame address & bit to be injected  
                for item in rand_items:
                    bitIndex = item[1]
                    frameAddress = item[2] 
                    
                    # Read a frame
                    frames = self._commandManager.ReadFrame(frameAddress, 1)
                    #Flip the value of the specific bit in the specific frame Address
                    frames[0].BitFlip(bitIndex)

                    #Inject the fault by writing the specific frame
                    status = self._commandManager.WriteFrame(frames)
                    Log.Print(str(status))

                #Try a range of diffrent inputs for one fault
                for j in range(DIFF_INPUTS_TO_TEST):

                    #Generate a random input number within the range
                    r=random.randint(0,4294967295)

                    #Pass the random input through BSCAN
                    status = self._commandManager.WriteBscanRegister(23,r)
                    Log.Print(str(status))

                    #Read the result through BSCAN
                    bscanRegister = self._commandManager.ReadBscanRegister(22)
                    if bscanRegister!=2*r:
                        print("Fault Detected. Result: "+str(bscanRegister)+"\n")
                        for item in rand_items:
                            bitIndex = item[1]
                            frameAddress = item[2] 
                            print("total_errors|Frame:"+str(hex(frameAddress))+"|Bit:"+str(hex(bitIndex))+"\n")
                        print("")
                        total_errors+=1
                        break
                        
                for item in rand_items:
                    bitIndex = item[1]
                    frameAddress = item[2] 
                    
                    # Read a frame
                    frames = self._commandManager.ReadFrame(frameAddress, 1)

                    #correct fault
                    frames[0].BitFlip(bitIndex)
                    status = self._commandManager.WriteFrame(frames)
                    Log.Print(str(status))

            print("AVF:"+str(total_errors/INJECTIONS_TO_PERFROM)+"\n")

        except Exception as e:
            Log.PrintException(f'FrmApplication.__BtnStartClicked: {str(e)}')
        self.Exit()

        ############################# END of fault injection procedure ################################
